/*****************************************************************************
**
**  File Name:          hex_prog.c
**  Author:             Dave Helland
**  Date:               3/21/01
**
**  Purpose/Description: program FLASH in
**         Atmel AMR microprocessors running :  Bootload.c
**
**  CSC:  MegaLoad
**
**  Public Data:
**
**  Private Data:
**
**  Functions:
**
**  Version/Revision History:
**
**
**  Version ID   Date      Author        SPCR No.    Description
**
**  1.0    5/05/04   Initial Release
**
**
***************************************************************************/

/*********************************************************/
/*                   I N C L U D E                       */
/*********************************************************/
#include <stdio.h>
#include <ctype.h>

#ifdef __WIN32__
#include <conio.h>
#endif


#ifdef  __linux__
#include <unistd.h>
#include <fcntl.h>
#include "linux_serial.h"
#include "global.h"
#include "linux_screen.h"
#ifndef STDIN_FILENO
  #define STDIN_FILENO 0
#endif
#endif
/*********************************************************/
/*           P R O T O T Y P E                           */
/*********************************************************/
#include "defines.h"
#include "interface.h"
#include "main.h"
#include "hex_prog.h"
/*********************************************************/
//#define DEBUG_MEM  1



/*********************************************************/
/*           G L O B A L    V A R I A B L E S            */
/*********************************************************/
//int eep_rom = 0;
unsigned int avr_mem_max;
unsigned int avr_eeprom_max;


unsigned char DeviceID;
unsigned char FlashSize;
unsigned char BootSize;
unsigned char PageSize;
unsigned int PageByte;
unsigned char EEpromSize;
int i_PageSize;
MEMIDX_T i_ProgSize;
MEMIDX_T i_EepromSize;
int debug = 0;

//extern int did_flash;
char msgstring[80];


static MEMIDX_T mem_lo = 0;
static MEMIDX_T mem_hi = 0;
extern P_MEM mem_buf;
extern int verbose;


//This is the source to the HEX2BIN utility.  It's job is to
//conver and intel HEX file into a binary file.  The binary file
//is a perfect image of the program to be loaded into a ROM.

// From HEX2BIN V1.1 by Craig Pataky
// Modified 4/24/99 by David Helland


#define ERR_NONE        0
#define ERR_CHECKSUM    1
#define ERR_BADFILENAME 2
#define ERR_OVERFLOW    3
#define ERR_UNKNOWN     0xFF

#define WORD unsigned int
#define BYTE unsigned char
#define BOOL int
#define TRUE 1
#define FALSE 0

BYTE gError = ERR_NONE;
BYTE HUGE *Binary;

/*********************************************************/
/*                   C O D E                             */
/*********************************************************/

BYTE
atoh(BYTE a)
{
  a = (BYTE) toupper(a);
  if (a <= '9')
    a = (BYTE) (a - '0');
  else
    a = (BYTE) (10 + (a - 'A'));
  return (a);
}

BOOL
ReadRecord(int fd, BYTE * Buf, int max_len)
{
  BYTE l, c;

  c = 0;
  l = 0;

  while (c != 10) {
    if (read(fd, &c, 1) != 1)
      return (FALSE);
    if ((c != 13) && (c != ' ')) {
      Buf[l] = c;
      l++;
      if (l >= max_len) {
        cprintf("WARNING: file not HEX format!\n\r");
        gError = ERR_UNKNOWN;
       return (-1);
      }
    }
  }
  return (TRUE);
}

//TODO: Validate checksum!!!
BYTE
TranslateRecord(BYTE * Record, BYTE HUGE * Buf, MEMIDX_T * Address)
{
  BYTE b, k, j, len;
  BYTE csum, type;
  WORD addr;
  static long Ext_addr = 0;
  csum = 0;

  //len=0;
  if (Record[0] != ':') {
    cprintf("WARNING: file not HEX format!\r\n");
    gError = ERR_UNKNOWN;
    len = 0;
    return (len);
  }
  len = atoh(Record[1]);
  len = (BYTE) ((len << 4) + atoh(Record[2]));
  csum += len;

  addr = 0;
  addr = (addr << 4) + atoh(Record[3]);
  addr = (addr << 4) + atoh(Record[4]);
  csum += (BYTE) addr;
  addr = (addr << 4) + atoh(Record[5]);
  addr = (addr << 4) + atoh(Record[6]);
  csum += (BYTE) (addr & 0xFF);
  *Address = addr + Ext_addr;

  type = 0;
  type = (BYTE) ((type << 4) + atoh(Record[7]));
  type = (BYTE) ((type << 4) + atoh(Record[8]));
  csum += type;

  if ((type == 0x02) || (type == 0x04)) // extended address record
  {
    Ext_addr = 0;
    Ext_addr = (Ext_addr << 4) + atoh(Record[9]);
    Ext_addr = (Ext_addr << 4) + atoh(Record[10]);
    Ext_addr = (Ext_addr << 4) + atoh(Record[11]);
    Ext_addr = (Ext_addr << 4) + atoh(Record[12]);
    if (type == 0x02)
      Ext_addr = Ext_addr << 4;
    if (type == 0x04)
      Ext_addr = Ext_addr << 16;
  }

  k = 0;
  for (j = 0; j < len; j++) {
    b = 0;
    b = (BYTE) ((b << 4) + atoh(Record[k + 9]));
    k++;
    b = (BYTE) ((b << 4) + atoh(Record[k + 9]));
    k++;
    Buf[j] = b;
    csum += b;
  }


  b = 0;
  b = (BYTE) ((b << 4) + atoh(Record[k + 9]));
  k++;
  b = (BYTE) ((b << 4) + atoh(Record[k + 9]));
  csum += b;

  if (csum != 0) {
    if (gError != ERR_NONE)
      cprintf("WARNING: HEX file corrupt!\r\n");
    gError = ERR_CHECKSUM;
  }
  if (type != 0)
    len = 0;
  return (len);
}





BOOL
TranslateFile(char *HexFile)
{
  BYTE Buffer[100];
  BYTE Record[100];
  MEMIDX_T HiAddr;
  MEMIDX_T LoAddr;
  MEMIDX_T Address;
  BYTE Length;
  int j, fd, result;


  Binary = (BYTE HUGE *) mem_buf;


  HiAddr = 0;
  LoAddr = 0x1ffff;

//  fd = open(HexFile, O_BINARY | O_RDONLY);
  fd = open(HexFile, O_RDONLY);
  if (fd <= 0) {
    gError = ERR_BADFILENAME;
    return (FALSE);
  }
  while (1) {

    result = ReadRecord(fd, Record,100);
    if (gError != ERR_NONE) {
      close(fd);
      return (FALSE);
    }
    if (result == FALSE) break;

    //     cprintf("File: %s\r\n", Record);        TEST for HEX FILE !!!!!!!!
    Length = TranslateRecord(Record, Buffer, &Address);
    if (gError != ERR_NONE) {
      close(fd);
      return (FALSE);
    }
    if (Length != 0) {
      for (j = 0; j < Length; j++) {
        if ((Address + j) > (MEM_MAX * 2 + 1)) {
          gError = ERR_OVERFLOW;
          close(fd);
          return (FALSE);
        }
        Binary[Address + j] = Buffer[j];
        if ((Address + j) > HiAddr)
          HiAddr = Address + j;
        mem_hi = (HiAddr >> 1) + 1;
        if ((Address + j) < LoAddr)
          LoAddr = Address + j;
        mem_lo = (LoAddr >> 1);
      }
    }
  }  // end of  while (ReadRecord(fd, Record,100))
//  if (HiAddr!=0) swab(Binary, Binary, HiAddr);
  cprintf("LoAddr =  %x    HiAddr =  %x\r\n", mem_lo, mem_hi);
  close(fd);
  return (TRUE);
}



static int
read_hex_file(char *hexfile)
{
  if (!TranslateFile(hexfile)) {
    if (gError == ERR_OVERFLOW)
      cprintf("ERROR: HEX file too large!\r\n");
    if (gError == ERR_CHECKSUM)
      cprintf("ERROR: HEX file corrupt!\r\n");
    if ((gError == ERR_UNKNOWN) || (gError == ERR_NONE))
      cprintf("ERROR: Cannot translate file!\r\n");
    if (gError == ERR_BADFILENAME)
      cprintf("ERROR: Could not find HEX file!\r\n");
    my_exit("<<TRANSLATION ABORTED>>\r\n");
  }

  cprintf("Translation complete!\r\n");
  return 1;
}


#if 0
static char *
my_gets(char *buf, int len)
{
  char *p;

  *buf = 0;
  if (fgets(buf, len - 1, stdin) == NULL)
    return NULL;
  buf[len - 1] = '\0';
  if ((p = strchr(buf, '\r\n')) != NULL)
    *p = '\0';
  if ((p = strchr(buf, '\r')) != NULL)
    *p = '\0';
  return buf;
}

#endif



int
read_next_ch(BYTE * ch)
{
  unsigned char ichar[1];
  int i;

  for (i = 0; i < 1000; i++) {
    if (ChkSerialInput() > 0) {
      ReadSerialChar(ichar);
      *ch = ichar[0];
//     cprintf("char received = %c\r\n",*ch);
      return 1;
    }
    my_delay(1);
  }
  return 0;
}


int
wait_for_ch(char exp_ch)
{
  BYTE ch;
  if (read_next_ch(&ch)) {
//     cprintf("char received = %x\r\n",ch);
    if (ch == exp_ch)
      return 1;
    else
      return 0;
  }
  return 0;
}


int
Get_Device_Id(int mega_type)
{

  switch (mega_type) {
  case 8:
    DeviceID = 'A';             // Mega8
    FlashSize = 'l';            // Flash 8k
    PageSize = 'R';             // 64 Bytes
    PageByte = 64;              // 64 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  case 16:
    DeviceID = 'B';             // Mega16
    FlashSize = 'm';            // Flash 16k
    PageSize = 'S';             // 128 Bytes
    PageByte = 128;             // 128 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  case 32:
    DeviceID = 'E';             // Mega32
    FlashSize = 'n';            // Flash 32k
    PageSize = 'S';             // 128 Bytes
    PageByte = 128;             // 128 Bytes
    EEpromSize = '2';           // 1024 Byte
    break;

  case 64:
    DeviceID = 'C';             // Mega64
    FlashSize = 'o';            // Flash 64k
    PageSize = 'T';             // 256 Bytes
    PageByte = 256;             // 256 Bytes
    EEpromSize = '3';           // 2048 Byte
    break;

  case 128:
    DeviceID = 'D';             // Mega128
    FlashSize = 'p';            // Flash 64k
    PageSize = 'T';             // 256 Bytes
    PageByte = 256;             // 256 Bytes
    EEpromSize = '4';           // 4096 Byte
    break;

  case 162:
    DeviceID = 'F';             // Mega162
    FlashSize = 'm';            // Flash 16k
    PageSize = 'S';             // 128 Bytes
    PageByte = 128;             // 128 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  case 169:
    DeviceID = 'G';             // Mega169
    FlashSize = 'm';            // Flash 16k
    PageSize = 'S';             // 128 Bytes
    PageByte = 128;             // 128 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  case 8515:
    DeviceID = 'H';             // Mega8515
    FlashSize = 'l';            // Flash 8k
    PageSize = 'R';             // 64 Bytes
    PageByte = 64;              // 64 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  case 8535:
    DeviceID = 'I';             // Mega8535
    FlashSize = 'l';            // Flash 8k
    PageSize = 'R';             // 64 Bytes
    PageByte = 64;              // 64 Bytes
    EEpromSize = '1';           // 512 Byte
    break;

  default:
    my_exit(" Invalid Mega Type\r\n");
  }
  return 0;
}


int
Set_Device_Id(void)
{

  switch (DeviceID) {
  case 'A':
    if (verbose)
      cprintf("DeviceID = Mega8\r\n");
    break;
  case 'B':
    if (verbose)
      cprintf("DeviceID = Mega16\r\n");
    break;
  case 'E':
    if (verbose)
      cprintf("DeviceID = Mega32\r\n");
    break;
  case 'C':
    if (verbose)
      cprintf("DeviceID = Mega64\r\n");
    break;
  case 'D':
    if (verbose)
      cprintf("DeviceID = Mega128\r\n");
    break;
  case 'F':
    if (verbose)
      cprintf("DeviceID = Mega162\r\n");
    break;
  case 'G':
    if (verbose)
      cprintf("DeviceID = Mega169\r\n");
    break;
  case 'H':
    if (verbose)
      cprintf("DeviceID = Mega8515\r\n");
    break;
  case 'I':
    if (verbose)
      cprintf("DeviceID = Mega8535\r\n");
    break;
  default:
    my_exit("DeviceID unknown");
  }

  switch (FlashSize) {
  case 'l':
    if (verbose)
      cprintf("FlashSize = 8k\r\n");
    i_ProgSize = 8192 - BOOTSTRAP;
    break;
  case 'm':
    if (verbose)
      cprintf("FlashSize = 16k\r\n");
    i_ProgSize = 16384 - BOOTSTRAP;
    break;
  case 'n':
    if (verbose)
      cprintf("FlashSize = 32k\r\n");
    i_ProgSize = 32768 - BOOTSTRAP;
    break;
  case 'o':
    if (verbose)
      cprintf("FlashSize = 64k\r\n");
    i_ProgSize = 65536 - BOOTSTRAP;
    break;
  case 'p':
    if (verbose)
      cprintf("FlashSize = 64k\r\n");
    i_ProgSize = 65536 - BOOTSTRAP;
    break;
  default:
    my_exit("FlashSize unknown");
  }

  switch (PageSize) {
  case 'R':
    if (verbose)
      cprintf("PageSize = 64 bytes\r\n");
    i_PageSize = 64;
    break;
  case 'S':
    if (verbose)
      cprintf("PageSize = 128 bytes\r\n");
    i_PageSize = 128;
    break;
  case 'T':
    if (verbose)
      cprintf("PageSize = 256 bytes\r\n");
    i_PageSize = 256;
    break;
  default:
    my_exit("PageSize unknown\r\n");
  }


  switch (EEpromSize) {
  case '1':
    if (verbose)
      cprintf("EEpromSize = 512 bytes\r\n");
    i_EepromSize = 512;
    break;
  case '2':
    if (verbose)
      cprintf("EEpromSize = 1024 bytes\r\n");
    i_EepromSize = 1024;
    break;
  case '3':
    if (verbose)
      cprintf("EEpromSize = 2048 bytes\r\n");
    i_EepromSize = 2048;
    break;
  case '4':
    if (verbose)
      cprintf("EEpromSize = 4096 bytes\r\n");
    i_EepromSize = 4096;
    break;
  default:
    my_exit("EEpromSize unknown\r\n");
  }
  return 1;
}



int burn_flash(int megatype, int verify)
{
  int i;
  int j;
  MEMIDX_T addr;
  MEMIDX_T temp, save;
//    MEMIDX_T page_addr;
//    MEMIDX_T lo = 0, hi;
  unsigned short val, retry;
  unsigned char adr_hi, adr_lo, val_hi, val_lo;
//    unsigned int page_start;
  unsigned int page_wordsz;
//    char *rom_name;
//    unsigned int page_buf[128];
//    unsigned int pg_mask,pg_shift;
  unsigned short chk_sum_1, chk_sum_2;
  unsigned char in_csum[4];
  unsigned char out_csum[4];
  unsigned char in_char;
  int err_cnt = 0;
//    int page;
#ifdef DEBUG_MEM
  int line_len;
#endif
//    page = 0;

  Get_Device_Id(megatype);
  Set_Device_Id();

  if ((mem_hi) > i_ProgSize) {
    my_exit("burn_flash() Flash address too big...");
  }

  //i = 0;
#if 0
  page_start = mem_lo;

  if (i_PageSize == 64) {
    pg_mask = 0x001f;
    pg_shift = 5;
  }                             // 32 words
  if (i_PageSize == 128) {
    pg_mask = 0x003f;
    pg_shift = 6;
  }                             // 64 words
  if (i_PageSize == 256) {
    pg_mask = 0x007f;
    pg_shift = 7;
  }                             // 128 words
  if ((page_start & pg_mask) != 0) {
    my_exit("\r\nabort: address must be on a page boundry");
  }
#endif
  page_wordsz = i_PageSize >> 1;

  //  for (j = 0; j < (int)page_wordsz; j++) page_buf[j] = 0xffff;
  addr = mem_lo;
  retry = 0;

  while (addr <= mem_hi) {      // these are WORD addresses
    save = addr;

retry_write:
#ifdef DEBUG_MEM
    line_len = 0;
#endif
    /* convert address to word address */
    //temp = addr >> 1;
    temp = addr;
    /* get hi and lo parts of page byte number */
    adr_hi = (BYTE) (((unsigned int) temp >> 8) & 0xff);
    adr_lo = (BYTE) ((unsigned int) temp & 0xff);
    /* write word address to AVR */
  try_address:
    chk_sum_1 = (short) addr;
    chk_sum_2 = (short) addr;
    WriteSerialChar('A');
    WriteSerialChar((unsigned char) adr_hi);
    WriteSerialChar((unsigned char) adr_lo);
    if (!read_next_ch(&in_char)) {
      my_exit("\r\nabort: read char time-out flash word program completion");
    }
    if (in_char != '\r') {
      if (retry++ > 2) {
        my_exit("\r\nabort: flash address retry count exhausted");
      }
      err_cnt++;
      goto try_address;
    }

    WriteSerialChar('D');       // non-standard ultra fast load
    if (verify) {
      WriteSerialChar('l');     // checksum
      WriteSerialChar('l');     // ored with checksum (could be 0)
    }
    else {
      WriteSerialChar('k');     // checksum
      WriteSerialChar('k');     // ored with checksum (could be 0)
    }
    /* write a page of data to the AVR */
    for (j = 0; j < (int) page_wordsz; j++) {
      /* get the WORD value from HUGE array */
      if (addr > mem_hi)
        val = 0xffff;
      else
        val = mem_buf[addr++];
      /* convert to BYTE values */
      val_hi = (BYTE) ((val >> 8) & 0xff);
      val_lo = (BYTE) (val & 0xff);
      /* If write mode -------write hi and lo bytes to AVR */
      if (!verify) {
        WriteSerialChar((unsigned char) val_lo);
        WriteSerialChar((unsigned char) val_hi);
      }
      //cprintf("val sent %04x \r\n", val);
      /* add data values to BYTE checksum */
      chk_sum_1 += val;
      chk_sum_2 ^= val;

    /* write BYTE checksum to AVR */
    val = chk_sum_1;
    /* convert to BYTE values */
    out_csum[0] = (BYTE) ((val >> 8) & 0xff);
    out_csum[1] = (BYTE) (val & 0xff);
    val = chk_sum_2;
    /* convert to BYTE values */
    out_csum[2] = (BYTE) ((val >> 8) & 0xff);
    out_csum[3] = (BYTE) (val & 0xff);

#ifdef DEBUG_MEM
      {
        if (line_len == 0)
          cprintf("%04x: ", addr - 1);
        cprintf("%04x ", val);
        line_len++;
        if (line_len == 8) {
          line_len = 0;
          cprintf("\r\n");
        }
      }
#endif
    }   // end of write page of data to AVR

    /* write hi and lo bytes to AVR */
    if (!verify) {
      for (i = 0; i < 4; i++)
        WriteSerialChar((unsigned char) out_csum[i]);
      if (verbose) cprintf("sent page: %u\r\n", save);
    }
    else if (verbose) cprintf("check page: %u\r\n", save);

    if (verify) {
      for (i = 0; i < 4; i++) {
        if (!read_next_ch(&in_csum[i])) {
          cprintf("\r\nabort: read char time-out flash verify completion\r\n");
          return 0;             // still want exitbootif()
        }
      }
      for (i = 0; i < 4; i++) {
        if (in_csum[i] != out_csum[i]) {
          err_cnt++;
          cprintf("\r\nerror at page: %u\r\n", save);
          if (verbose) {
            cprintf("csum sent = ");
            for (i = 0; i < 4; i++) cprintf("%2x ",out_csum[i]);
            cprintf("\r\ncsum recv = ");
            for (i = 0; i < 4; i++) cprintf("%2x ",in_csum[i]);
            cprintf("\r\n");
          }
          if (err_cnt > 5) {
            cprintf("too many errors, bailing out\r\n");
            return 0;           // still want exitbootif()
          }
          addr = save;
          if (verbose) cprintf("retry count =%d \r\n", err_cnt);
          goto retry_write;
        } // checksum failure
      }
    }                           // end of verify

    else {
      /* wait for results of write character */
      if (!read_next_ch(&in_char)) {
        my_exit("\r\nabort: read char time-out flash word program completion\r\n");
      }
      /* check for failed checksum indication */
      if (in_char == 'N') {
#if 0
        for (i = 0; i < 4; i++) {
          if (!read_next_ch(&in_csum[i])) {
            cprintf("abort: read char time-out flash verify completion\r\n");
            return 0;             // still want exitbootif()
          }
        }
        cprintf("csum sent = ");
        for (i = 0; i < 4; i++) cprintf("%2x ",out_csum[i]);
        cprintf("\r\ncsum recv = ");
        for (i = 0; i < 4; i++) cprintf("%2x ",in_csum[i]);
        cprintf("\r\n");
#endif

        if (retry++ > 4) {
          my_exit("\r\nabort: flash programming retry count exhausted");
        }
        addr = save;
        cprintf("retry count =%d \r\n", retry);
        err_cnt++;
        goto retry_write;
      }
      /* check for OK to send anaother page character */
      if (in_char != '\r') {
        cprintf("\r\nfail OK to send anaother page character\r\n");
        cprintf("abort: wrong character received\r\n");
        sprintf(msgstring,"got 0x%x (%c)  wanted (0x0d) \r\n", in_char, in_char);
        my_exit(msgstring);
      }
      //   cprintf("got %c\r\n",in_char);
      //   cprintf("Sent page: %d\r\n",page);
    }

    cprintf(".");
    //cprintf("page: %d\r\n",page);
    //page++;
    fflush(stdout);
  }                             // while addr <= mem_hi

  if (verbose) cprintf("\r\nend WORD address = %x\r\n", addr);

  if (verify) {
    if (verbose)
      cprintf("\r\nTOTAL serious errors during verify: %u\r\n", err_cnt);
    if (err_cnt == 0)
      cprintf("\r\nSUCCESS\r\n");
    else
      cprintf("\r\nFAILURE\r\n");
  }
  else {
    if (verbose)
      cprintf("\r\nTOTAL serious errors during write: %u\r\n", err_cnt);
  }
  return 1;
}







int
burn_eeprom(int mode, int megatype)
{                               /* add BF and BE support */
//    long i;
  int i;
  MEMIDX_T addr;
  MEMIDX_T temp, save;
//    MEMIDX_T page_addr;
//    MEMIDX_T lo = 0, hi;
  unsigned int val, retry;
  unsigned char adr_hi, adr_lo, val_hi, val_lo;
//    char *rom_name;
//    unsigned int page_buf[128];
  unsigned char chk_sum;
  unsigned char in_char;
//    int page;
#ifdef DEBUG_MEM
  int line_len = 0;
#endif
//    page = 0;


  Get_Device_Id(megatype);
  Set_Device_Id();

  if (mode != ROM_EEPROM)
    goto skipit;

  //  check for byte address too big for EEPROM
  if ((mem_hi << 1) > i_EepromSize) {
    my_exit("burn_eeprom() EEPROM address too big...");
  }

  addr = mem_lo;

  while (addr <= mem_hi) {      // these are WORD addresses
    save = addr;

    for (i = 0; i < 2; i++) {
      chk_sum = 0;
      retry = 0;

    retry_eeprom:
      /* EEP_ROM format is one byte per WORD address in mem_buf,
         otherwise, hex format has 2 bytes per WORD location */
      if (!eep_rom)             // then must be WORD addresses
      {
        temp = (addr << 1) + i; // convert to byte address
      }
      /* get hi and lo parts of page byte number */
      adr_hi = (BYTE) (((unsigned int) temp >> 8) & 0xff);
      adr_lo = (BYTE) ((unsigned int) temp & 0xff);
      /* write page byte number to AVR */
      WriteSerialChar((unsigned char) adr_hi);
      chk_sum += (BYTE) adr_hi;
      WriteSerialChar((unsigned char) adr_lo);
      chk_sum += (BYTE) adr_lo;

      /* get the WORD value from HUGE array */
      val = mem_buf[addr];
      /* convert to BYTE values */
      val_hi = (BYTE) ((val >> 8) & 0xff);
      val_lo = (BYTE) (val & 0xff);
      /* write hi and lo bytes to AVR */
      val = val_lo;
      if ((i == 1) && (!eep_rom))
        val = val_hi;
      WriteSerialChar((unsigned char) val);
      chk_sum += (BYTE) val;

#ifdef DEBUG_MEM
      {
        if (line_len == 0)
          cprintf("%x: ", temp);
        cprintf("%02x ", val);
        line_len++;
        if (line_len == 16) {
          line_len = 0;
          cprintf("\r\n");
        }
      }
#endif

      /* write BYTE checksum to AVR */
      WriteSerialChar((unsigned char) chk_sum);
      /* wait for results of write character */
      if (!read_next_ch(&in_char)) {
        my_exit
          ("abort: read char time-out on eeprom byte program completion");
      }
      /* check for failed checksum indication */
      if (in_char == '@') {
        if (retry++ > 4) {
          my_exit
            ("abort: programming retry count exhausted eeprom byte program");
        }
        addr = save;
        cprintf("retry count =%d \r\n", retry);
        goto retry_eeprom;
      }
      /* check for OK to send anaother page character */
      if (in_char != '!') {
        my_exit("abort: eeprom write - wrong character received");
      }
      //   cprintf("got %c\r\n",in_char);
      if (eep_rom)
        i++;                    // then must be 1 byte per WORD addresses
    }                           // for i = 0 to 1    i.e. write 2 bytes from memory words
    addr++;

    cprintf(".");
    //cprintf("page: %d\r\n",page);
    //page++;
    fflush(stdout);
  }                             // while addr <= mem_hi

  cprintf("\r\nend address = %x\r\n", temp);

skipit:
  /*  tell AVR we are finished EEPROM programming  */
  WriteSerialChar((unsigned char) 0xff);
  WriteSerialChar((unsigned char) 0xff);

  return 1;
}



int
read_rom_file(int rom_type, char *filename)
{
  char hex_str[80];
  int linenum = 0;
  int i, val, len;
  MEMIDX_T addr;
  MEMIDX_T lo = 0;
  MEMIDX_T hi;
  int first_time = 1;
  unsigned int adr_hi, adr_lo, val_hi, val_lo;
  char *p;
  FILE *in;

  switch (rom_type) {
  case ROM_FLASH:
    hi = avr_mem_max;
    len = 11;
    break;
  case ROM_EEPROM:
    hi = avr_eeprom_max;
    len = 7;
    break;
  default:
    sprintf(msgstring,"Error! read_rom_file: bad type (0x%x)!\r\n", rom_type);
    my_exit(msgstring);
  }

  if (read_hex_file(filename))
    return 1;

  if ((in = fopen(filename, "rt")) == NULL) {
    //perror(filename);
    sprintf(msgstring,"Cannot open input file %s!\r\n", filename);
    my_exit(msgstring);
  }
  mem_lo = mem_hi = 0;
  while (!feof(in)) {
    fgets(hex_str, 20, in);
    if ((p = strchr(hex_str, '\n')) != NULL)
      *p = '\0';
    if ((p = strchr(hex_str, '\r')) != NULL)
      *p = '\0';
    ++linenum;
    if ((int) strlen(hex_str) != len) {
      cprintf("Bad data at %s line %u\r\n", filename, linenum);
      cprintf("   Bad input : %s\r\n", hex_str);
      cprintf("   Bad len   : %u (expected %d)\r\n", strlen(hex_str), len);
      return 0;
    }
    if (debug)
      cprintf("%s \r\n", hex_str), fflush(stdout);
    switch (rom_type) {
    case ROM_FLASH:
      i = sscanf(hex_str, "%4x%2x:%x", &adr_hi, &adr_lo, &val);
      break;
    case ROM_EEPROM:
      i = sscanf(hex_str, "%2x%2x:%x", &adr_hi, &adr_lo, &val);
      eep_rom = 1;
      break;
    }
    if (i != 3) {
      cprintf("Bad format at %s line %u (%d)\r\n", filename, linenum, i);
      cprintf("   Bad input : %s\r\n", hex_str);
      return 0;
    }
    addr = (adr_hi << 8) | adr_lo;
    val_lo = val & 0xff;
    val_hi = (val >> 8) & 0xff;
    if (addr > hi) {
      cprintf("Bad address at %s line %u\r\n", filename, linenum);
      cprintf("   Bad input : %s\r\n", hex_str);
      cprintf("   Bad addr  : 0x%x (expected %u...%u)\r\n", addr, lo, hi);
      return 0;
    }
    mem_buf[addr] = (BYTE) (val_hi << 8) | (BYTE) val_lo;
    if (first_time) {
      mem_lo = mem_hi = (unsigned int) addr;
      first_time = 0;
    }
    if (addr < mem_lo)
      mem_lo = (unsigned int) addr;
    if (addr > mem_hi)
      mem_hi = (unsigned int) addr;
  }
  return 1;
}



