/*
I think we can whip that up tonight - I have a Linux machine with a spare serial port
we can test with.  Here's some stuff for the basic characteristics settings - it's for
multiple OS'es so you can see how things map.  On Linux you can do "man tcsetattr" and
it'll show all the options for tcsetattr() and related routines - which are what you
want and what are used in this module.  For reading characters on Linux you can just
use the low-level read() function, and the open can be done with:

    FILE *fp = fopen("/dev/ttyS0","w") ;
    int fd = fileno(fp) ;

kind of thing.  Or you can use the low-level open():

    int fd = open("/dev/ttyS0",O_RDWR) ;

either way.  For writing you would use "write(fd,buffer,len)" and it should return
"len" if all went well...

That should get ya started...
*/

typedef struct  {
    ULONGWORD   ulBaudRate ;            /* true baud rate in bps */
    UCHAR       ucDataBits ;            /* # of data bits 5/6/7/8 */
    UCHAR       ucStopBits ;            /* # of stop bits 1/2 */
    UCHAR       ucParity ;              /* parity - as below */

#define CSDEV_PARITY_NONE   0
#define CSDEV_PARITY_ODD    1
#define CSDEV_PARITY_EVEN   2
#if 0
#define CSDEV_PARITY_MARK   3
#define CSDEV_PARITY_SPACE  4
#endif

    UCHAR       ucFlowControl ;         /* flow control - as below */

#define CSDEV_FLOW_NONE     0
#define CSDEV_FLOW_XONXOFF  1
#define CSDEV_FLOW_RTSCTS   2

} CSDEV_COMM_ATTRS ;

RETCODE
rCsSetDeviceCommAttrs(
    CS_FILE fDevice,
    CSDEV_COMM_ATTRS *pAttrs
){
    IO_CONTROL  *pIoControl ;

    if ( (pIoControl = CsIoGetIOC(fDevice)) == NULL) {
        return RETURN_CODE(CS_STATUS_ERROR, CS_IOERR_INVALID_ARCS_NUM) ;
    }
#if defined(WNT)
#define GENERATED
    {
    DCB         CommDCB ;
    HANDLE      hDevice ;
    hDevice = pIoControl->hRead ;
    CommDCB.DCBlength = sizeof CommDCB ;
    if (GetCommState(hDevice,&CommDCB))  {

        CommDCB.fBinary = TRUE ;
        CommDCB.fAbortOnError = FALSE ;
        CommDCB.fOutxDsrFlow = FALSE ;
        CommDCB.fDsrSensitivity = FALSE ;
        CommDCB.fDtrControl = DTR_CONTROL_ENABLE ;
        CommDCB.fRtsControl = RTS_CONTROL_ENABLE ;
        CommDCB.fOutxCtsFlow = FALSE ;
        CommDCB.fOutX = FALSE ;
        CommDCB.fInX = FALSE ;

        CommDCB.BaudRate = pAttrs->ulBaudRate ;

        CommDCB.ByteSize = pAttrs->ucDataBits ;

        CommDCB.Parity = pAttrs->ucParity ;

        CommDCB.StopBits = pAttrs->ucStopBits&2 ;   /* 0/2 == 1/2 stopbits */

        switch (pAttrs->ucFlowControl)  {
            case CSDEV_FLOW_NONE:  {
                break ;
            }
            case CSDEV_FLOW_RTSCTS:  {
                CommDCB.fOutxCtsFlow = TRUE ;
                CommDCB.fRtsControl = RTS_CONTROL_HANDSHAKE ;
                break ;
            }
            case CSDEV_FLOW_XONXOFF:  {
                CommDCB.fInX = TRUE ;
                CommDCB.fOutX = TRUE ;
                CommDCB.XonChar = 0x11 ;
                CommDCB.XoffChar = 0x13 ;
                CommDCB.XonLim = 10 ;
                CommDCB.XoffLim = 50 ;
                break ;
            }
        }

        SetCommState(hDevice,&CommDCB) ;
    }
    return(SERVSTAT_TO_RETCODE(GetLastError())) ;
    }
#endif

#if defined(ARCS_POSIX_1)
#define GENERATED
/*
 * for QNX systems - needs more work for other non-linuxes probably.
 */
#if !defined(CRTSCTS)
#define CRTSCTS (IHFLOW|OHFLOW)
#endif
    {
    struct termios TermIo ;
    speed_t Speed ;
    int     fd ;

    fd = pIoControl->Fd ;
    if (tcgetattr(fd,&TermIo) != -1)  {

        TermIo.c_iflag &= ~(INPCK|IXON|IXOFF) ;
        TermIo.c_iflag |=  (IGNPAR) ;

        TermIo.c_cflag &= ~(CSIZE|CSTOPB|PARENB|PARODD|CRTSCTS) ;

        switch (pAttrs->ulBaudRate)  {
            case 110:       {  Speed = B110 ; break ; }
            case 300:       {  Speed = B300 ; break ; }
            case 600:       {  Speed = B600 ; break ; }
            case 1200:      {  Speed = B1200 ; break ; }
            case 2400:      {  Speed = B2400 ; break ; }
            case 4800:      {  Speed = B4800 ; break ; }
            default:
                /* fall through */
            case 9600:      {  Speed = B9600 ; break ; }
            case 19200:     {  Speed = B19200 ; break ; }
            case 38400:     {  Speed = B38400 ; break ; }
#if defined(B57600)
            case 57600:     {  Speed = B57600 ; break ; }
#endif
#if defined(B115200)
            case 115200:    {  Speed = B115200 ; break ; }
#endif
#if defined(B230400)
            case 230400:    {  Speed = B230400 ; break ; }
#endif
        }
        cfsetospeed(&TermIo,Speed) ;
        cfsetispeed(&TermIo,Speed) ;

        switch (pAttrs->ucDataBits)  {
            case 5:  { TermIo.c_cflag |= CS5 ; break ; }
            case 6:  { TermIo.c_cflag |= CS6 ; break ; }
            case 7:  { TermIo.c_cflag |= CS7 ; break ; }
            default:
                /* fall through */
            case 8:  { TermIo.c_cflag |= CS8 ; break ; }
        }

        switch (pAttrs->ucParity)  {
            case CSDEV_PARITY_ODD:  {
                TermIo.c_cflag |= (PARENB|PARODD) ;
                break ;
            }
            case CSDEV_PARITY_EVEN:  {
                TermIo.c_cflag |= (PARENB) ;
                break ;
            }
        }

        if (pAttrs->ucStopBits == 2)  {
            TermIo.c_cflag |= CSTOPB ;
        }

        switch (pAttrs->ucFlowControl)  {
            case CSDEV_FLOW_NONE:  {
                break ;
            }
            case CSDEV_FLOW_RTSCTS:  {
                TermIo.c_cflag |= CRTSCTS ;
                break ;
            }
            case CSDEV_FLOW_XONXOFF:  {
                TermIo.c_iflag |= (IXON|IXOFF) ;
                break ;
            }
        }

        if (tcsetattr(fd,TCSANOW,&TermIo) != -1)  {
            return(CS_RETCODE_SUCCESS) ;
        }
    }
    return(SERVSTAT_TO_RETCODE(errno)) ;
    }
#endif

#if !defined(GENERATED)
    return(RETURN_CODE(CS_STATUS_FATAL,CS_REASON_UNSPECIFIED)) ;
#endif

}


}
 