//---------------------------------------------------------------------------
// Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
//  TODO.C - COM functions required by MLANLL.C, MLANTRNU, MLANNETU.C and
//           MLanFile.C for MLANU to communicate with the DS2480 based
//           Universal Serial Adapter 'U'.  Fill in the platform specific code.
//
//  Version: 1.02
//
//  History: 1.00 -> 1.01  Added function msDelay.
//
//           1.01 -> 1.02  Changed to generic OpenCOM/CloseCOM for easier
//                         use with other platforms.
//

//--------------------------------------------------------------------------
// Copyright (C) 1998 Andrea Chambers and University of Newcastle upon Tyne,
// All Rights Reserved.
//--------------------------------------------------------------------------
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE UNIVERSITY OF NEWCASTLE UPON TYNE OR ANDREA CHAMBERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//---------------------------------------------------------------------------
//
//  LinuxLNK.C - COM functions required by MLANLLU.C, MLANTRNU.C, MLANNETU.C
//             and MLanFile.C for MLANU to communicate with the DS2480 based
//             Universal Serial Adapter 'U'.  Platform specific code.
//
//  Version: 1.03
//  History: 1.00 -> 1.03 modifications by David Smiczek
//                        Changed to use generic OpenCOM/CloseCOM
//                        Pass port name to OpenCOM instead of hard coded
//                        Changed msDelay to handle long delays
//                        Reformatted to look like 'TODO.C'
//                        Added #include "ds2480.h" to use constants.
//                        Added function SetBaudCOM()
//                        Added function msGettick()
//                        Removed delay from WriteCOM(), used tcdrain()
//                        Added wait for byte available with timeout using
//                          select() in ReadCOM()

/* Mods by S. J. Melhuish

   getlock for use by OpenCOM and CloseCOM
   tries to open / close a lock on the device fd
*/


#ifndef __WIN32__

#include <stdio.h>
#include <unistd.h>
//#include <sys/types.h>
//#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <time.h>
#include <termios.h>
#include <errno.h>
#include <sys/time.h>
#include <stdarg.h>

//#include "config.h"

//#include "ds2480.h"

/* Error reporting - SJM */
#include "linux_serial.h"
#include "werr.h"

/* For applctn_quick_poll() */
//#include "applctn.h"

 FILE *fp;

/* Which branch is in use? */
extern int devices_active_branch ;

struct termios origterm;

/* Added by SJM */
long csGettick(void);

// Exportable functions for opening/closing serial port
//int OpenCOM(int, char *);
//void CloseCOM(int);

// LinuxLNK global
int fd;
int werr_message_level = werr_level_errors_and_warnings ;
int debug_level = werr_level_debug_0 ;

//--------------------------------------------------------------------------
// Write an array of bytes to the COM port, verify that it was
// sent out.  Assume that baud rate has been set.
//
// Returns 1 for success and 0 for failure
//
int WriteCOM(int dummyport, int outlen, unsigned char *outbuf)
{
   long count = outlen;
   int i ;

   //FlushCOM(dummyport); /* Dump pending RX / TX */
//   i = write(fd, outbuf, outlen);
   for (i=0; i<outlen;i++)  fputc((int)outbuf[i],fp);
  fflush(fp);

//   tcdrain(fd);
//   if (i!=count) werr(WERR_DEBUG0, "WriteCOM failed");
   return (i == count);
}

//--------------------------------------------------------------------------
// Read number of bytes the COM port has waiting to be read
//
// The number of characters to be read
// Returns 0, or 1
//
int TstSerialInput(int dummyport)
{
  fd_set         filedescr;
  struct timeval tval ;

    // set a descriptor to wait for a character available
    FD_ZERO(&filedescr);
    FD_SET(fd,&filedescr);
    /* set timeout to 10ms - minimum value */
    tval.tv_sec = 0;
    tval.tv_usec = 0;
//    tval.tv_usec = 0;

    // if byte available read or return bytes read
    if (select(fd+1,&filedescr,NULL,NULL,&tval) != 0)  return 1;
    return 0;
}

//--------------------------------------------------------------------------
// Read an array of bytes to the COM port, verify that it was
// sent out.  Assume that baud rate has been set.
//
// The number of characters read is written into nread
// Returns 0, or -1 on error
//
int ReadCOM(int dummyport, int inlen, unsigned char *inbuf, int *nread)
{
  fd_set         filedescr;
  struct timeval tval ;
//  long           start_tick, elapsed = 0L ;
  long           elapsed = 0L ;

//  start_tick = csGettick() ; /* Will use for 2-s timeout */
  *nread = 0 ;

  // loop to wait until each byte is available and read it
  while (*nread < inlen)
  {
    // set a descriptor to wait for a character available
    FD_ZERO(&filedescr);
    FD_SET(fd,&filedescr);
    /* set timeout to 0ms - use TstSerialInput first */
    tval.tv_sec = 0;
    tval.tv_usec = 0;

    // if byte available read or return bytes read
    if (select(fd+1,&filedescr,NULL,NULL,&tval) != 0)
    {
      if (read(fd,&inbuf[*nread],1) == 1) ++*nread ;
    }
    else
    {
      return 0;
#if 0
      elapsed = csGettick() - start_tick ;
      /* Timeout on select */
      /* Check to see if it's time to give up */
      if (elapsed >= 200L)
      {
        werr(WERR_DEBUG0, "ReadCOM 2-s timeout") ;
        return -1 ;
      }
      else
      {
        /* Don't give up yet, but let the GUI poll whilst we're waiting */
        applctn_quick_poll(0) ;
      }
#endif
    }
  }

  if (elapsed) werr(WERR_DEBUG0, "ReadCOM took >= %d cs", (int) elapsed) ;

  // success, so return desired length
  return 0 ;
}


//---------------------------------------------------------------------------
//  Description:
//     flush the rx and tx buffers
//
void FlushCOM(int dummyport)
{
   tcflush (fd, TCIOFLUSH);
}


//--------------------------------------------------------------------------
//  Description:
//     Delay for at least 'len' ms
//
#if 1
void msDelay(int len)
{

   usleep(1000 * (unsigned long) len) ;
#if 0
   struct timespec s;              // Set aside memory space on the stack

   s.tv_sec = len / 1000;
   s.tv_nsec = (len - (s.tv_sec * 1000)) * 1000000;
   nanosleep(&s, NULL);
#endif
}
#endif

#if 0
void msDelay(int len)
{
  struct timeval tv, end_tv ;
  struct timezone tz ;

  /* For short delays, just sleep */
  if (len <= 10)
  {
    usleep(1000UL * (unsigned long) len) ;
    return ;
  }

  /* For longer delays we'll do some polling whilst we wait */
  /* Get start time */
  gettimeofday(&tv, &tz) ;

  /* Calculate end time */
  end_tv.tv_sec = tv.tv_sec + len / 1000 ;
  end_tv.tv_usec = (len - (tv.tv_sec * 1000)) * 1000 ;

  /* Loop whilst waiting */
  do
  {
    /* Quick poll */
//    applctn_quick_poll(0) ;

    /* Wait for 10 ms */
    usleep(10000UL) ;

    /* Check time now */
    gettimeofday(&tv, &tz) ;
  } while
    ((tv.tv_sec < end_tv.tv_sec) ||
     ((tv.tv_sec == end_tv.tv_sec) && (tv.tv_usec < end_tv.tv_usec))) ;
}
#endif
//--------------------------------------------------------------------------
//  Description:
//     Send a break on the com port for at least 2 ms
//
void BreakCOM(int dummyport)
{
   int duration = 0;              // see man termios break may be
   tcsendbreak(fd, duration);     // too long

   /* The trunk will be active following reset */
//   devices_active_branch = -1 ;
}


static int getlock(int type)
{
#  ifdef ENABLE_SERIALLOCK
#  ifdef __linux__
  /* F_SETLK for linux only (?) */
   struct flock lockinfo ;
   /* fd is global */

   /* Try for a lock */
   lockinfo.l_whence = SEEK_SET ;
   lockinfo.l_start  = 0 ;
   lockinfo.l_len    = 0 ;
   lockinfo.l_type   = type ;

   return (fcntl(fd, F_SETLK, &lockinfo)) ;
#  else
   return 0 ;
#  endif // ifdef __linux__
#  else
   return 0 ;
#  endif /* ENABLE_SERIALLOCK */
}


//---------------------------------------------------------------------------
// Attempt to open a com port.
// Set the starting baud rate to 9600.
//
// 'port_zstr' - zero terminate port name.  Format is platform
//               dependent.
//
// Returns: TRUE - success, COM port opened
//

int OpenCOM(int dummyport, char *port_zstr)
{
   struct termios t;               // see man termios - declared as above
   int rc;

   werr(WERR_DEBUG0, "Opening %s", port_zstr) ;
   fd = open(port_zstr, O_RDWR | O_NONBLOCK | O_NOCTTY);
   if (fd<0) 
   {
     werr(WERR_DEBUG0, "Opening %s failed", port_zstr) ;
     return fd;
   }

   fp = fdopen(fd, "w");

   werr(WERR_DEBUG0, "Obtaining lock") ;
   /* Try for a write lock on the COM port */
   if (getlock(F_WRLCK) /*flock(fd, LOCK_EX)*/ < 0)
   {
     /* Didn't get the lock - abort */
     werr(WERR_WARNING, "Unable to obtain lock on device %s", port_zstr) ;
     close(fd) ;
     fd = -1 ;
     return 0 ;
   }
   werr(WERR_DEBUG0, "Have lock") ;

   rc = tcgetattr (fd, &t);
   if (rc < 0)
   {
      int tmp;
      werr(WERR_WARNING, "Failed to get serial port attributes") ;
      tmp = errno;
      close(fd);
      errno = tmp;
      return rc;
   }

   cfsetospeed(&t, B9600);
   cfsetispeed (&t, B9600);

   tcgetattr(fd,&t);
   // Save original settings.
   origterm = t;

   // Set to non-canonical mode, and no RTS/CTS handshaking
   t.c_iflag &= ~(BRKINT|ICRNL|IGNCR|INLCR|INPCK|ISTRIP|IXON|IXOFF|PARMRK);
   t.c_iflag |= IGNBRK|IGNPAR;
   t.c_oflag &= ~(OPOST);
   t.c_cflag &= ~(CRTSCTS|CSIZE|HUPCL|PARENB);
   t.c_cflag |= (CLOCAL|CS8|CREAD);
   t.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON|IEXTEN|ISIG);
   t.c_cc[VMIN] = 0;
//   t.c_cc[VTIME] = 3;
   t.c_cc[VTIME] = 0;

   rc = tcsetattr (fd, TCSAFLUSH, &t);
   tcflush(fd, TCIOFLUSH);

   if (rc < 0)
   {
      int tmp;
      werr(WERR_WARNING, "Failed to set serial port attributes") ;
      tmp = errno;
      close(fd);
      errno = tmp;
      return rc;
   }

   return fd;
}


//---------------------------------------------------------------------------
// Closes the connection to the port.
//
void CloseCOM(int dummyport)
{
   if (fd < 0) return ;

   // restore tty settings
   tcsetattr(fd, TCSAFLUSH, &origterm);

   FlushCOM(dummyport);

   /* Release our lock */
   getlock(F_UNLCK) /*flock(fd, LOCK_UN)*/ ;
   close(fd);
}


//--------------------------------------------------------------------------
// Set the baud rate on the com port.
//
int SetBaudCOM(int dummyport, int new_baud)
{
   struct termios t;
   int rc;
   speed_t baud = B9600 ;

   // read the attribute structure
   rc = tcgetattr(fd, &t);
   if (rc < 0)
   {
      close(fd);
      return rc;
   }

   // convert parameter to linux baud rate
   switch(new_baud)
   {
            case 110:       {  baud = B110 ; break ; }
            case 300:       {  baud = B300 ; break ; }
            case 600:       {  baud = B600 ; break ; }
            case 1200:      {  baud = B1200 ; break ; }
            case 2400:      {  baud = B2400 ; break ; }
            case 4800:      {  baud = B4800 ; break ; }
            case 9600:      {  baud = B9600 ; break ; }
            case 19200:     {  baud = B19200 ; break ; }
            case 38400:     {  baud = B38400 ; break ; }
#if defined(B57600)
            case 57600:     {  baud = B57600 ; break ; }
#endif
#if defined(B115200)
            case 115200:    {  baud = B115200 ; break ; }
#endif
#if defined(B230400)
            case 230400:    {  baud = B230400 ; break ; }
#endif
            default:        {  baud = B9600 ; break ; }
   }

   // set baud in structure
   cfsetospeed(&t, baud);
   cfsetispeed(&t, baud);

   // change baud on port
   rc = tcsetattr(fd, TCSAFLUSH, &t);
   if (rc < 0) {
      close(fd);
      return rc;
   }
   return 1;
}


//--------------------------------------------------------------------------
// Get the current millisecond tick count.  Does not have to represent
// an actual time, it just needs to be an incrementing timer.
//
long msGettick(void)
{
   struct timezone tmzone;
   struct timeval  tmval;
   long ms;
   static time_t st = 0 ;

   if (st == 0) time(&st) ;

   gettimeofday(&tmval,&tmzone);
   ms = ((tmval.tv_sec - st) & 0x1FFFFF) * 1000 + tmval.tv_usec / 1000;
   return ms;
}


//--------------------------------------------------------------------------
// Get the current centisecond tick count.  Does not have to represent
// an actual time, it just needs to be an incrementing timer.
//
long csGettick(void)
{
   struct timezone tmzone;
   struct timeval  tmval;
   long cs;
   static time_t st = 0 ;

   if (st == 0) time(&st) ;

   gettimeofday(&tmval,&tmzone);
   cs = ((tmval.tv_sec - st) & 0xFFFFFF) * 100 + tmval.tv_usec / 10000;
   return cs;
}

// ============================================================

void Serial_SetDTR (void) {
   int set;
   ioctl (fd, TIOCMGET, &set);
   set |= TIOCM_DTR;
   ioctl (fd, TIOCMSET, &set);
}


void Serial_ClearDTR (void) {
   int set;
   ioctl (fd, TIOCMGET, &set);
   set &= ~TIOCM_DTR;
   ioctl (fd, TIOCMSET, &set);
}

static void werr_zap_nl(char *string, int limit)
{
  int i ;

  for (i=0; i<limit; ++i)
  {
    switch (string[i])
    {
      case '\n':
      case '\r':
        string[i] = ' ' ;
        break ;

      case '\0':
        return ;
    }
  }
}




static int werr_check_level(int flags, int level)
{
  switch(level)
  {
    case werr_level_none:
      return 0 ;

    case werr_level_fatal_only:
      return (flags & WERR_FATAL) ;

    case werr_level_errors:
      return (flags < WERR_WARNING) ;

    case werr_level_errors_and_warnings:
      return (flags <= WERR_WARNING) ;

    case werr_level_debug_0:
      return (flags <= WERR_DEBUG0) ;

    case werr_level_debug_1:
      return (flags <= WERR_DEBUG1) ;

    case werr_level_debug_2:
      return (flags <= WERR_DEBUG2) ;

    case werr_level_debug_3:
      return (flags <= WERR_DEBUG3) ;
  }

  return 1 ;
}

static int
werr_will_display(int flags)
{
  /* A fatal error must always call werr, so that it will die */
  return
    ((flags & WERR_FATAL) || werr_check_level(flags, werr_message_level)) ;
}

static int
werr_will_write(int flags)
{
  return ( werr_check_level(flags, debug_level)) ;
}

int
werr_will_output(int flags)
{
  return (werr_will_display(flags) || werr_will_write(flags)) ;
}

void werr(int flags, char* format, ...)
{
  #define WTS 256
   char task_name[] = "Serial I/O" ;
   //_kernel_oserror er;
   va_list va;
   char title[128], timestamp[64] ;
   char message[2048] ;
   time_t t ;
   //int max_flags ;


   if (!werr_will_output(flags)) return ;

   //er.errnum = 0;
   va_start(va, format);
   vsprintf(message, format, va);
   va_end(va);

   //strncpy(er.errmess, message, 251) ;

   time(&t) ;
   strftime(timestamp, 63, "%d, %H:%M:%S", localtime(&t)) ;
   sprintf(title, "Warning from %s [%s]", task_name, timestamp) ;


  /* Check reporting level */
  if (werr_check_level(flags, werr_message_level))
  {

    werr_zap_nl(message, sizeof(message)) ;

    fprintf(stderr, "%s\n", message) ;
  }
  return;
}

#endif  //#ifndef __WIN32__


