/*************************************************************************/
/* VTDECODE.C - VT100 decoder.                                           */
/*                                                                       */
/* Changes : 8/4/99 DRH - Shorten for 2 x 40 LCD panel                   */
/* Changes : 1/22/03 DRH - Compatible for PC screen                      */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*************************************************************************/

#ifdef __WIN32__
#include <conio.h>
#endif

#ifdef __linux__
#include <ncurses.h>
#endif

#include <stdio.h>
#include <ctype.h>

#include "defines.h"
#include "interface.h"

#define	VT_ROWS	24				/* # of rows */
#define	VT_COLS	80				/*  and columns in a screen */
#define	VT_NARGS	2			/* max args in ESC. seq. */
#define	VT_OUTBUF_SIZE	512			/* max output size */

#define	SET	0				/* mode SET */
#define	RESET	1				/* mode RESET */

#define	VS_CHAR		0
#define	VS_ESCSEQ	1
#define	VS_CHARSET	2
#define	VS_CSIQUES	3

#define	ESC	27
#define	CSI	'['
#define	DO_LINEFEED 1
#define	CTL(_c)  ((_c)-'@')	   /* control character */

typedef unsigned char  u08;
typedef          char  s08;
typedef unsigned short u16;
typedef          short s16;


int vt_gotoxy( void );
void vt_clreol( void );
void vt_rcvalid( void );
void vt_erase_screen( void );
void vt_erase_line( void );
void vt_linefeed( void );
void do_cr( void );
void do_lf( void );

int first_time = 1;

char state = VS_CHAR;		  /* current state */
char func_mod;		  /* function modifier */
int row = 0 ;		  /* current row mem offset from mem_top */
int col = 0 ;		  /*  and column */
int last_ch = 0 ;		  /* last character seen */
int args[VT_NARGS] ;	  /* argument list */
int arg  = 0 ;		  /* current argument index */

//  args[0] = 0 ;	/* reset it */

/**********************************/
/*                                */
/* VTDECODE.C - VT decoder.       */
/*                                */
/**********************************/

int
vtdecode( unsigned char rx_ch ) 	/* char to decode */

{
  unsigned char ch ;			/* current character */

     ch = rx_ch ;	   /* grab the character */


     switch (state)  		/* dispatch according to state */
       {

/*
* VS_CHAR - normal characters.
*
* includes processing for control characters as well as printables.
*/
       case VS_CHAR:  {			/* normal character */

output_char: ;
	 switch (ch)  {			/* dispatch on character */
	   case ESC:  {			/* ESCape */
	     state = VS_ESCSEQ ;	/* change state */
	     break ;
	   }
	   case CTL('H'): {		/* Control-H (backspace) */
	     if (col > 0) col--;
             my_putchar(ch);
	     break ;
	   }
#ifdef DO_LINEFEED
	   case CTL('J'):  {  		/* linefeed */
	     vt_linefeed() ;		/* do linefeed */
	     break ;
	   }
	   case CTL('K'): {		/* Vertical Tab */
	       vt_linefeed() ;		/* do linefeed */
	       break ;
	   }
	   case CTL('L'):  {		/* Form Feed */
	       vt_linefeed() ;		/* do linefeed */
	       break ;
	   }
#endif
	   case CTL('M'):  {  		/* Carriage Return */
	     col = 0 ;			/* reset column */
#ifdef __WIN32__
             my_putchar(ch);
#endif
#ifdef __linux__
             do_cr();
#endif
	     break ;			/* we done */
	   }
	   default:  {			/* normal character */
             if (ch != 0x09) {
	       if (ch < ' ')  {		/* if control character */
	         goto done ;			/* then ignore it */
	       }
             }
	     /* ------------------ normal char write ----------------*/
             my_putchar(ch);
	   } /* default */
	 } /* switch on char ch */
	 break ;
       } /* case VS_CHAR */
/*
* VS_ESCSEQ - processing an escape sequence.
*/
       case VS_ESCSEQ:  {

/*
* first check for ESC-x sequences - most of which we just ignore.
*/
	 if (last_ch == ESC)  {	/* if just after an escape */

	   switch (ch)  {		/* dispatch on character */

#ifdef DO_LINEFEED
	     case 'E':  {		/* ESC-E - CR-LF */
	       //show_cmd("ESC-E - CR-LF");
	       col = 0 ;		/* go to left margin */
	       vt_linefeed() ;		/* do linefeed */
	       break ;
	     }

	     case 'D':  {		/* ESC-D - linefeed */
	       //show_cmd("ESC-D - linefeed");
	       vt_linefeed() ;		/* do linefeed */
	       break ;
	     }
#endif

	     case CSI:  {		/* Control Seq. Introducer */
	       arg = 0 ;		/* reset arg index */
	       args[0] = 0 ;    	/* reset it */
	       last_ch = CSI ;	        /* last was CSI */
	       goto done ;		/* handle more stuff */
	     }

	   } /* switch ch after ESC */
	   state = VS_CHAR ;	/* revert to character mode */

	 } /* if last ch = ESC */

	 else  {			/* not just after escape */
/*
* if we get here, then we are processing something in the standard form
* of ESC-CSI...
*/
	   if (isdigit(ch))  {		/* if it's a digit */
	     args[arg] = args[arg]*10+(ch-'0') ;
	     goto done ;
	   }
	   else  {
	     switch (ch)  {		/* dispatch on char */

	       case ESC:  {		/* another sequence */
		 goto output_char ;	/* abort current one for now */
	       }
	       case ';':  {		/* argument separator */
		 if (++(arg) >= VT_NARGS)  {	/* if too many */
		   arg = VT_NARGS-1 ;	/* reset to end of list */
		 }
		 else  {			/* in range */
		   args[arg] = 0 ;	/* clear it out */
		 }
		 goto done ;
	       }



	       case 'f':		/* ESC [ f - HVP */
	       case 'H':  {		/* ESC [ H - CUP */
		 //show_cmd("cursor position");
		 //vt_setarg(2) ;		/* set up for two args */
		 row = args[0]-1 ;	/* set row */
		 col = args[1]-1 ;	/*  and column */
		 vt_rcvalid() ;		/* make sure valid */
                 //printf("(%i,%i)",col+1,row+1);
#ifdef __linux__
		 gotoxy(col+1,row+1) ;
#endif
		 break ;
	       }




	       case 'J':  {		/* ESC [ J - Erase */
		 state = VS_CHAR ;	/* get into character mode */
		 //show_cmd("erase screen");
		 vt_erase_screen();
		 break ;
	       }

	       case 'K':  {		/* ESC [ K - Erase in line */
		 state = VS_CHAR ;	/* get into character mode */
		 //show_cmd("erase line");
		 vt_erase_line();
		 break ;
	       }


	     } /* switch on ch */
	     state = VS_CHAR ;	/* back to character mode */
	   }  /* else not digit */
	 }  /* else not last ESC */
	 break ;
       } /* case VS_ESC_SEQ */


     }  /* end of switch on state */
done:
     last_ch = ch ;		/* remember last character */

   return(1) ;
}

#ifdef __linux__
void
do_cr( void )
{
  int x,y;

  getyx(stdscr,y,x);
  move(y,0);
}

void
do_lf( void )  // fix curses problem of erasing to eol before line feed
{
  int x,y,max_x,max_y;
  getmaxyx(stdscr,max_y,max_x);
  getyx(stdscr,y,x);
  if (y + 1 == max_y) scrl(1);
  else move(y+1,x);
}
#endif

/**********************************/
/*                                */
/* vt_linefeed   - move down 1    */
/*                    line        */
/**********************************/

void
vt_linefeed( void )
{
#ifdef __WIN32__
   my_putchar(0x0a);
#endif
#ifdef __linux__
//   my_putchar(0x0d);
   do_lf();
#endif
//   putchar(0x0a);
   row++ ;
   vt_rcvalid();
//   gotoxy(col+1,row+1) ;
  return;
}


/**********************************/
/*                                */
/* VT_CLREOL   - clear to end of  */
/*                   current line */
/**********************************/

void
vt_clreol( void )
{
  int i;
   for (i = col; i < VT_COLS; i++)
      {
#ifdef __linux__
      gotoxy(i+1,row+1);
#endif
      my_putchar(' ');
      }
#ifdef __linux__
   gotoxy(col+1,row+1) ;
#endif
   return;
}






/**********************************/
/*                                */
/* VT_RCVALID.C - Make sure       */
/*          row/column are valid. */
/*                                */
/**********************************/

void
vt_rcvalid( void )

{
   if (row < 0)  row = 0 ;
   if (row > VT_ROWS-1)  row = VT_ROWS-1 ;
   if (col < 0)  col = 0 ;
   if (col > VT_COLS-1)  col = VT_COLS-1;
   return;
}






/**********************************/
/*                                */
/* VT_ERASE_SCREEN                */
/*                                */
/**********************************/

void
vt_erase_screen( void )

{

#ifdef __linux__
	if (args[0] == 2)	/* entire window */
        clrscr();
#endif
   return;
}  /* end vt_erase_screen */







/**********************************/
/*                                */
/* VT_ERASE_LINE                  */
/*                                */
/**********************************/

void
vt_erase_line( void )

{
  int i;
  int temp_col;

   switch (args[0])  {	/* dispatch on function */
     case 0:  {		/* cursor to end of line */
#ifdef __linux__
		 clreol();
       gotoxy(col+1,row+1) ;
#endif
       break ;
     }
     case 1:  {		/* beg. line -> cursor */
	for (i=0; i <= col; i++)
	   {
#ifdef __linux__
		   gotoxy(i+1,row+1);
#endif
           my_putchar(' ');
	   }
#ifdef __linux__
	gotoxy(col+1,row+1) ;
#endif
	break;
     }
     case 2:  {		/* entire line */
	temp_col = col;
	col = 0;
#ifdef __linux__
	clreol();
#endif
	col = temp_col;
#ifdef __linux__
	gotoxy(col+1,row+1) ;
#endif
       break;
     }
   }
   return;
}

#if 0
void main(void)
{
char ch;
  row = 0 ;		/* reset row */
  col = 0 ;		/* reset col */
  arg = 0 ;		/* reset arg index */
  args[0] = 0 ;	/* reset it */
  last_ch = 0 ;	        /* last was CSI */
  state = VS_CHAR;
  ch = ESC;
  vtdecode(ch);
}
#endif





